
import numpy as np
import math
from scipy.sparse import csr_matrix, diags
from scipy import linalg
import time

try:
  from numba import jit, njit
  print("module 'numba' is installed")
  numbaOn = True
except ModuleNotFoundError:
  numbaOn = False
  print("module 'numba' not installed")

if numbaOn:
  @njit(["void(float64[:], f8, float64[:], float64[:], f8, f8)"])
  def velocityImplNumba(u, t, f, expVec, dxInvHalf, mu0):
    n = len(u)
    uSq = np.square(u)
    f[0] = dxInvHalf * (math.pow(mu0, 2) - uSq[0]) + expVec[0]
    for i in range(1,n):
      f[i] = dxInvHalf * ( uSq[i-1] - uSq[i] ) + expVec[i]
else:
  def velocityImplNumba(u, t, f, expVec, dxInvHalf, mu0):
    n = len(u)
    uSq = np.square(u)
    f[0] = dxInvHalf * (math.pow(mu0, 2) - uSq[0]) + expVec[0]
    for i in range(1,n):
      f[i] = dxInvHalf * ( uSq[i-1] - uSq[i] ) + expVec[i]

if numbaOn:
  @njit(["void(float64[:], float64[:], float64[:], f8)"])
  def fillDiag(u, diag, ldiag, dxInv):
    n = len(u)
    for i in range(n-1):
      diag[i] = -dxInv*u[i]
      ldiag[i] = dxInv*u[i]
    diag[n-1] = -dxInv*u[n-1]
else:
  def fillDiag(u, diag, ldiag, dxInv):
    n = len(u)
    for i in range(n-1):
      diag[i] = -dxInv*u[i]
      ldiag[i] = dxInv*u[i]
    diag[n-1] = -dxInv*u[n-1]

class Burgers1d:
  def __init__(self, Ncell):
    self.mu_    = np.array([5., 0.02, 0.02])
    self.xL_    = 0.
    self.xR_    = 100.
    self.Ncell_ = Ncell
    self.dx_    = 0.
    self.dxInv_ = 0.
    self.dxInvHalf_ = 0.
    self.xGrid_ = np.zeros(self.Ncell_)
    self.U0_    = np.zeros(self.Ncell_)
    self.expVec_= np.zeros(self.Ncell_)
    self.diag_  = np.zeros(self.Ncell_)
    self.ldiag_ = np.zeros(self.Ncell_-1)
    self.setup()

  def setup(self):
    self.dx_ = (self.xR_ - self.xL_)/float(self.Ncell_)
    self.dxInv_ = (1.0/self.dx_)
    self.dxInvHalf_ = 0.5 * self.dxInv_
    for i in range(0, self.Ncell_):
      self.U0_[i] = 1.
      self.xGrid_[i] = self.dx_*i + self.dx_*0.5
    self.expVec_ = self.mu_[1] * np.exp( self.mu_[2] * self.xGrid_ )

  def createVelocity(self):
    return np.zeros(self.Ncell_)

  def velocity(self, u, t, f):
    velocityImplNumba(u, t, f[:], self.expVec_,
                      self.dxInvHalf_, self.mu_[0])

  def createApplyJacobianResult(self, B):
    return np.zeros_like(B)

  def applyJacobian(self, u, B, t, result):
    J = self.jacobian(u, t)
    result[:] = J.dot(B)

  def jacobian(self, u, t):
    fillDiag(u, self.diag_, self.ldiag_, self.dxInv_)
    return diags( [self.ldiag_, self.diag_], [-1,0], format='csr')




import numpy as np
from scipy import linalg
import matplotlib.pyplot as plt
import random

# if run from within a build of pressio4py, need to append to python path
import pathlib, sys
file_path = pathlib.Path(__file__).parent.absolute()
sys.path.append(str(file_path) + "/..")         # to access doFom

from pressio4py import rom as rom, logger
from pressio4py import solvers as solvers
from pressio4pyApps.advection_diffusion1d import AdvDiff1d
from adv_diff_1d_fom import doFom
from settings_for_website import edit_figure_for_web

#----------------------------
class MyLinSolver:
  def __init__(self):
    pass
  def solve(self, A,b,x):
    lumat, piv, info = linalg.lapack.dgetrf(A, overwrite_a=True)
    # here we must use x[:] otherwise it won't overwrite x passed in
    x[:], info = linalg.lapack.dgetrs(lumat, piv, b, 0, 0)

#----------------------------
class MyMasker:
  def __init__(self, indices):
    self.rows_ = indices
    self.sampleMeshSize_ = len(indices)

  def createApplyMaskResult(self, operand):
    if (operand.ndim == 1):
      return np.zeros(self.sampleMeshSize_)
    else:
      return np.zeros((self.sampleMeshSize_, operand.shape[1]))

  def applyMask(self, operand, time, result):
    if (operand.ndim == 1):
      result[:] = np.take(operand, self.rows_)
    else:
      result[:] = np.take(operand, self.rows_, axis=0)

#----------------------------------------
def computePodModes(snapshots):
  print("SVD on matrix: ", snapshots.shape)
  U,S,VT = np.linalg.svd(snapshots)
  return U

#----------------------------------------
def runMaskedGalerkin(fomObj, dt, nsteps, modes, sampleMeshIndices):
  # find out number of modes wanted
  romSize = modes.shape[1]

  # fom reference state: here it is zero
  fomReferenceState = np.zeros(fomObj.nGrid)

  # create ROM state by projecting the fom initial condition
  fomInitialState = fomObj.u0.copy()
  romState = np.dot(modes.T, fomInitialState)

  '''
  creating a masked Galerkin problem involves these steps:
  (1) creating the decoder on the FULL mesh
  (2) create a "projector operator" by filtering the rows
      of the POD modes only on the sample mesh (aka mask) indices.
      The projector is responsible to project the FOM velocity.
      Note that one can use other matrices for the projector
      but that will be shown in other demos.
  (3) create a masker object responsible to mask the FOM operators.
  (4) create the masked Galerkin problem
  '''

  # 1. create a linear decoder
  linearDecoder = rom.Decoder(modes)

  # 2. create the projector
  # here, simply use "collocation" with the POD modes filtered on the "sample mesh"
  modesOnSampleMesh = np.take(modes, sampleMeshIndices, axis=0)
  projector = rom.galerkin.ArbitraryProjector(modesOnSampleMesh)

  # 3. create the masker object
  masker = MyMasker(sampleMeshIndices)

  # 4. create the masked galerkin problem with Euler forward
  problem = rom.galerkin.masked.ProblemBackwardEuler(fomObj, linearDecoder,
                                                     romState, fomReferenceState,
                                                     masker, projector)

  # linear and non linear solver
  lsO = MyLinSolver()
  nlsO = solvers.NewtonRaphson(problem, romState, lsO)
  nlsO.setMaxIterations(15)

  # solve the problem
  rom.galerkin.advanceNSteps(problem, romState, 0., dt, nsteps, nlsO)

  # after we are done, use the reconstructor object to reconstruct the fom state
  # NOTE: even though the Galerkin problem was run on the "masked mesh points",
  # this reconstruction uses the POD modes on the full mesh stored in the decoder
  # so we can effectively obtain an approximation of the full solution
  fomRecon = problem.fomStateReconstructor()
  return [fomRecon.evaluate(romState), romState]

#----------------------------------------
def runMaskedLspg(fomObj, dt, nsteps, modes, sampleMeshIndices):
  # find out number of modes wanted
  romSize = modes.shape[1]

  # fom reference state: here it is zero
  fomReferenceState = np.zeros(fomObj.nGrid)

  # create ROM state by projecting the fom initial condition
  fomInitialState = fomObj.u0.copy()
  romState = np.dot(modes.T, fomInitialState)

  '''
  creating a masked LSPG problem involves these steps:
  (1) creating the decoder on the FULL mesh
  (2) create a masker object responsible to mask the FOM operators
  (3) create the masked LSPG problem
  '''

  # 1. create a linear decoder
  linearDecoder = rom.Decoder(modes)

  # 2. create the masker object
  masker = MyMasker(sampleMeshIndices)

  # 3. create the masked galerkin problem with Euler forward
  problem = rom.lspg.unsteady.masked.ProblemEuler(fomObj, linearDecoder,
                                                  romState, fomReferenceState,
                                                  masker)

  # linear and non linear solver
  lsO = MyLinSolver()
  nlsO = solvers.GaussNewton(problem, romState, lsO)
  nlsO.setMaxIterations(10)

  # solve the problem
  rom.lspg.solveNSequentialMinimizations(problem, romState, 0., dt, nsteps, nlsO)

  # after we are done, use the reconstructor object to reconstruct the fom state
  # NOTE: even though the Galerkin problem was run on the "masked mesh points",
  # this reconstruction uses the POD modes on the full mesh stored in the decoder
  # so we can effectively obtain an approximation of the full solution
  fomRecon = problem.fomStateReconstructor()
  return [fomRecon.evaluate(romState), romState]


######## MAIN ###########
if __name__ == "__main__":
  logger.initialize(logger.logto.terminal, "null")
  logger.setVerbosity([logger.loglevel.info])

  # total number of grid points
  meshSize = 200

  # create fom object
  fomObj = AdvDiff1d(nGrid=meshSize, adv_coef=1.0)

  # the final time to integrate to
  finalTime = .05

  #--- 1. FOM ---#
  fomTimeStepSize  = 1e-5
  fomNumberOfSteps = int(finalTime/fomTimeStepSize)
  sampleEvery      = 100
  [fomFinalState, snapshots] = doFom(fomObj, fomTimeStepSize, fomNumberOfSteps, sampleEvery)

  #--- 2. POD ---#
  modes = computePodModes(snapshots)

  #--- 3. MASKED GALERKIN and LSPG ROM ---#
  # a masked problem is supposed to make it easier to emulate the
  # effect of hyper-reduction. To create a mask ROM problem,
  # we need to select and provide to pressio a set of indices
  # identifying a subset of the grid points in the full mesh.
  # This is a simple way to mimic hyper-reduction
  # without changing the FOM problem. In fact, the fom still
  # computes the full operators but we have an additional step
  # to "mask" the operators to compute the sample mesh version.
  # In this test, the meshSize = 200. Our sample mesh includes
  # the two end points since those contain the boundary conditions,
  # and 20 randomly selected grid points inside the domain.
  # So effectively we use 1/10 of the full mesh.
  random.seed(22123)
  sampleMeshSize = 20
  sampleMeshIndices = random.sample(range(1, 199), sampleMeshSize)
  sampleMeshIndices = np.append(sampleMeshIndices, [0, 199])
  # sort for convenience, not necessarily needed
  sampleMeshIndices = np.sort(sampleMeshIndices)

  romSize = 5  # number of modes to use
  romTimeStepSize  = 1e-4
  romNumberOfSteps = int(finalTime/romTimeStepSize)

  # run the masked galerkin problem
  [approximatedStateGal, romGal] = runMaskedGalerkin(fomObj, romTimeStepSize,
                                                     romNumberOfSteps,
                                                     modes[:,:romSize],
                                                     sampleMeshIndices)
  # run the masked galerkin problem
  [approximatedStateLspg, romLspg] = runMaskedLspg(fomObj, romTimeStepSize,
                                                   romNumberOfSteps,
                                                   modes[:,:romSize],
                                                   sampleMeshIndices)

  # compute l2-error between fom and approximate state
  fomNorm = linalg.norm(fomFinalState)
  err1 = linalg.norm(fomFinalState-approximatedStateGal)
  print("Galerkin: final state relative l2 error: {}".format(err1/fomNorm))
  err2 = linalg.norm(fomFinalState-approximatedStateLspg)
  print("LSPG: final state relative l2 error: {}".format(err2/fomNorm))


  #----------------------------------------------#
  #--- plot solutions on sample and full mesh ---#

  gridSM = fomObj.xGrid[np.sort(sampleMeshIndices)]
  phiSM = np.take(modes[:,:romSize], sampleMeshIndices, axis=0)
  # approxStateSM = phi_sm * romState (no need for reference state since = 0)
  approxStateGalSM  = np.dot(phiSM, romGal)
  approxStateLspgSM = np.dot(phiSM, romLspg)

  fig = plt.figure(0)
  ax = fig.gca()
  #plt.rcParams.update({'font.size': 18})
  ax.plot(fomObj.xGrid, fomFinalState, '-g', linewidth=1.5, label='FOM')

  ax.plot(gridSM, approxStateLspgSM, 'sy',
          markerfacecolor='None', markersize=7,
          label='LSPG solution on sample mesh')
  ax.plot(gridSM, approxStateGalSM, 'sm',
          markerfacecolor='None', markersize=7,
          label='Galerkin solution on sample mesh')
  ax.set_ylabel("Solution")
  ax.set_xlabel("x-coordinate")
  leg = plt.legend(fontsize=10, fancybox=True, framealpha=0, loc='lower right')
  ax.grid(True, linewidth=0.35, color='gray')
  #used to change color to text and axes
  edit_figure_for_web(ax, leg)
  plt.savefig('demo5_f1.png', dpi=250, transparent=True)
  plt.show()

  fig = plt.figure(1)
  ax1 = fig.gca()
  #plt.rcParams.update({'font.size': 18})
  ax1.plot(fomObj.xGrid, fomFinalState, '-g', linewidth=1.5, label='FOM')
  ax1.plot(fomObj.xGrid, approximatedStateLspg, 'oy',
          markerfacecolor='None', markersize=4,
          label='LSPG on full mesh')
  ax1.plot(fomObj.xGrid, approximatedStateGal, 'om',
          markerfacecolor='None', markersize=4,
          label='Galerkin solution on full mesh')

  ax1.set_ylabel("Solution")
  ax1.set_xlabel("x-coordinate")
  leg = plt.legend(fontsize=10, fancybox=True, framealpha=0, loc='lower right')
  ax1.grid(True, linewidth=0.35, color='gray')

  #used to change color to text and axes
  edit_figure_for_web(ax1, leg)
  plt.savefig('demo5_f2.png', dpi=250, transparent=True)
  plt.show()




import numpy as np
from scipy import linalg
import matplotlib.pyplot as plt
import sklearn.decomposition as skd

# if run from within a build of pressio4py, need to append to python path
import pathlib, sys
file_path = pathlib.Path(__file__).parent.absolute()
sys.path.append(str(file_path) + "/..")         # to access doFom

from pressio4py import rom as rom, logger
from pressio4py import solvers as solvers
from pressio4pyApps.advection_diffusion1d import AdvDiff1d
from adv_diff_1d_fom import doFom
from settings_for_website import edit_figure_for_web

class MyMapperKPCA:
  def __init__(self, snapshots, numModes):
    self.transformer_ = skd.KernelPCA(n_components=numModes,\
                                      kernel='poly',
                                      degree=3,
                                      fit_inverse_transform=True)
    # do training using provided snapshots
    self.transformer_.fit(snapshots)

    self.numModes_ = numModes
    fomSize = snapshots.shape[1]
    self.fomState0 = np.zeros(fomSize)
    self.fomState1 = np.zeros(fomSize)
    # attention: the jacobian of the mapping must be column-major oder
    # so that pressio can view it without deep copying it, this enables
    # to keep only one jacobian object around and to call the update
    # method below correctly
    self.jacobian_ = np.zeros((fomSize,numModes), order='F')

  def jacobian(self): return self.jacobian_

  def applyMapping(self, romState, fomState):
    fomState[:] = np.squeeze(self.transformer_.inverse_transform(romState.reshape(1,-1)))

  def applyInverseMapping(self, fomState):
    return np.squeeze(self.transformer_.transform(fomState.reshape(1,-1)))

  def updateJacobian(self, romState):
    romStateLocal = romState.copy()
    # finite difference to approximate jacobian of the mapping
    self.applyMapping(romStateLocal,self.fomState0)
    eps = 0.001
    for i in range(self.numModes_):
        romStateLocal[i] += eps
        self.applyMapping(romStateLocal, self.fomState1)
        self.jacobian_[:,i] = (self.fomState1 - self.fomState0) / eps
        romStateLocal[i] -= eps

#----------------------------------------
def runLspg(fomObj, dt, nsteps, customMapper):
  # this is an auxiliary class that can be passed to solve
  # LSPG to monitor the rom state.
  class RomStateObserver:
    def __init__(self): pass
    def __call__(self, timeStep, time, state): pass

  # this linear solver is used at each gauss-newton iteration
  class MyLinSolver:
    def __init__(self): pass
    def solve(self, A,b,x):
      lumat, piv, info = linalg.lapack.dgetrf(A, overwrite_a=True)
      x[:], info = linalg.lapack.dgetrs(lumat, piv, b, 0, 0)

  #----------------------------------------
  # create a custom decoder using the mapper passed as argument
  customDecoder = rom.Decoder(customMapper, "kPCAMapper")

  # fom reference state: here it is zero
  fomReferenceState = np.zeros(fomObj.nGrid)

  # create ROM state by projecting the fom initial condition
  fomInitialState = fomObj.u0.copy()
  romState = customMapper.applyInverseMapping(fomInitialState)

  # create LSPG problem
  problem = rom.lspg.unsteady.default.ProblemEuler(fomObj, customDecoder, romState, fomReferenceState)

  # create the Gauss-Newton solver
  nonLinSolver = solvers.GaussNewton(problem, romState, MyLinSolver())
  # set tolerance and convergence criteria
  nlsTol, nlsMaxIt = 1e-7, 10
  nonLinSolver.setMaxIterations(nlsMaxIt)
  nonLinSolver.setStoppingCriterion(solvers.stop.whenCorrectionAbsoluteNormBelowTolerance)

  # create object to monitor the romState at every iteration
  myObs = RomStateObserver()
  # solve problem
  rom.lspg.solveNSequentialMinimizations(problem, romState, 0., dt, nsteps, myObs, nonLinSolver)

  # after we are done, use the reconstructor object to reconstruct the fom state
  # get the reconstructor object: this allows to map romState to fomState
  fomRecon = problem.fomStateReconstructor()
  return fomRecon.evaluate(romState)

######## MAIN ###########
if __name__ == "__main__":
  logger.initialize(logger.logto.terminal, "null")
  logger.setVerbosity([logger.loglevel.info])

  # create fom object
  fomObj = AdvDiff1d(nGrid=120, adv_coef=2.0)

  # the final time to integrate to
  finalTime = .05

  #--- 1. FOM ---#
  fomTimeStepSize  = 1e-5
  fomNumberOfSteps = int(finalTime/fomTimeStepSize)
  sampleEvery      = 200
  [fomFinalState, snapshots] = doFom(fomObj, fomTimeStepSize, fomNumberOfSteps, sampleEvery)

  #--- 2. train a nonlinear mapping using kPCA ---#
  # here we use 3 modes, change this to try different modes
  myNonLinearMapper = MyMapperKPCA(snapshots.T, numModes=3)

  #--- 3. LSPG ROM ---#
  romTimeStepSize  = 3e-4
  romNumberOfSteps = int(finalTime/romTimeStepSize)
  approximatedState = runLspg(fomObj, romTimeStepSize, romNumberOfSteps, myNonLinearMapper)

  # compute l2-error between fom and approximate state
  fomNorm = linalg.norm(fomFinalState)
  err = linalg.norm(fomFinalState-approximatedState)
  print("Final state relative l2 error: {}".format(err/fomNorm))

  #--- plot ---#
  ax = plt.gca()
  plt.rcParams.update({'font.size': 18})
  ax.plot(fomObj.xGrid, fomFinalState, '-g', linewidth=2, label='FOM')
  ax.plot(fomObj.xGrid, approximatedState, 'or',
              markerfacecolor='None', markersize=5,
              label='LSPG: with kPCA mapping')
  ax.set_ylabel("Solution", fontsize=18)
  ax.set_xlabel("x-coordinate", fontsize=18)
  leg = plt.legend(fontsize=12, fancybox=True, framealpha=0, loc='lower right')
  ax.grid(True, linewidth=0.35, color='gray')

  #used to change color to text and axes
  edit_figure_for_web(ax, leg)
  plt.savefig('demo3.png', dpi=250, transparent=True)
  plt.show()




import numpy as np
import math
from scipy.sparse import csr_matrix, diags, spdiags
from scipy import linalg
import time

# def lindiff(u,x=None,dudx=None):
#   '''
#   So far only works with k(x) dependence
#   '''
#   k = 1. + 0.*u
#   dkdx = 0. + 0.*u
#   return k, dkdx

def nonlindiff(u,x=None,dudx=None):
  pw = 4.
  k = x**pw
  dkdx = pw*x
  return k,dkdx

class AdvDiff1d():
  '''
  du/dt = d/dx( k(u,x) du/dx ) - a*du/dx

  where the RHS is f and it will be approx by FD:

  d2u/dU2 = (u_{i+1} - 2u_{i} + u_{i-1})/dx^2
  du/dx = (u_{i} - u_{i-1})/dx

  with Dirichlet i.e. fixed (zero) boundary conditions

  For linear case, k(u,x) = 1
  For nonlinear case, k(u,x) = 1 - u^2
  (1-u^2)d2u/dx2 + du/dx(d/dx(1-u^2))
  = (1-u^2)*d2u/dx2 -2u*(du/dx)^2
  '''

  def __init__(self,
               kfun=nonlindiff,
               nGrid=50,
               xL=0.0, xR=1.0,
               IC=None,
               adv_coef=1.0):
    self.xL = xL
    self.xR = xR
    self.nGrid = nGrid+2 # include boundary points
    if IC==None:
      self.ic = lambda x: 2.*np.sin(9.*np.pi*x) - np.sin(4.*np.pi*x)
    else:
      self.ic = IC

    self.kfun = kfun
    self.adv_coef = adv_coef
    self.setup()

  def setup(self):
    self.xGrid = np.linspace(self.xL,self.xR,self.nGrid)
    self.dx = np.diff(self.xGrid)[0]

    if self.ic is None:
      self.ic = np.sin(np.pi*self.xGrid)
      self.u0 = self.ic.copy()
    else:
      self.u0 = self.ic(self.xGrid)

    # create 2nd order diff matrix
    self.diff_mtx = np.diag(-2.0*np.ones(self.nGrid), 0) + \
      np.diag(+1.0*np.ones(self.nGrid-1),-1) + \
      np.diag(+1.0*np.ones(self.nGrid-1),+1)
    self.diff_mtx[0] = np.zeros(self.nGrid)
    self.diff_mtx[-1] = np.zeros(self.nGrid)

    # create 1st order advection FD matrix
    self.adv_mtx = np.diag(1.0*np.ones(self.nGrid), 0) + \
      np.diag(-1.0*np.ones(self.nGrid-1), -1)
    self.adv_mtx[0] = np.zeros(self.nGrid)
    self.adv_mtx[-1] = np.zeros(self.nGrid)

  def createVelocity(self):
    return np.zeros(self.nGrid)

  def velocity(self, u, t, f):
    x = self.xGrid
    dx = self.dx
    A = self.diff_mtx
    d2udx2 = (1.0/dx**2)*np.dot(A,u)
    B = self.adv_mtx
    dudx = (1.0/dx)*np.dot(B,u)
    k, dkdx = self.kfun(u,x=x,dudx=dudx)
    a = self.adv_coef
    f[:] = k*d2udx2 + dudx*dkdx - a*dudx
    # enforce fixed boundary conditions
    f[0] = 0.
    f[-1] = 0.

  def createApplyJacobianResult(self, B):
    return np.zeros_like(B)

  def applyJacobian(self, u, B, t, result):
    J = self.jacobian(u, t)
    result[:] = np.dot(J,B)

  def jacobian(self, u, t):
    x = self.xGrid
    dx = self.dx
    A = self.diff_mtx
    B = self.adv_mtx
    a = self.adv_coef
    k, dkdx = self.kfun(u,x=x)
    J = np.dot(np.diag(k),(1./dx**2)*A) + \
      (1./dx**2)*np.dot(2*np.diag(dkdx*np.dot(B,u)),B) + \
      - a*(1./dx)*B
    return J




import numpy as np
import pathlib, sys

def doFom(fom, dt, nsteps, saveFreq):
  u = fom.u0.copy()
  U = [u]
  f = fom.createVelocity()
  for i in range(1,nsteps+1):
    # query rhs of discretized system
    fom.velocity(u, i*dt, f)
    # simple Euler forward
    u = u + dt*f
    if i % saveFreq == 0:
      U.append(u)
  Usolns = np.array(U)
  return [u, Usolns.T]



#!/usr/bin/env python
# Author:  Francesco Rizzi
# Contact: fnrizzi@sandia.gov, francesco.rizzi@ng-analytics.com

import os
import sys
import subprocess

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext

topdir = os.path.abspath(os.path.dirname(__file__))

def description():
    with open(os.path.join(topdir, 'DESCRIPTION.rst')) as f:
        return f.read()

# A CMakeExtension needs a sourcedir instead of a file list.
class CMakeExtension(Extension):
  def __init__(self, name, sourcedir=""):
    Extension.__init__(self, name, sources=[])
    self.sourcedir = os.path.abspath(sourcedir)

class CMakeBuild(build_ext):
  def build_extension(self, ext):
    extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))

    if not extdir.endswith(os.path.sep): extdir += os.path.sep

    if not os.path.exists(self.build_temp): os.makedirs(self.build_temp)
    print("self.build_temp ", self.build_temp)

    # CMake lets you override the generator - we need to check this.
    # Can be set with Conda-Build, for example.
    cmake_generator = os.environ.get("CMAKE_GENERATOR", "")

    #------------------------
    # --- pressio ---
    #------------------------
    pressioSrcDir   = ext.sourcedir + "/pressio"
    pressioBuildDir = self.build_temp+"/pressio/build"
    pressioPfxDir   = "../install"
    cmake_args = [
      "-DPRESSIO_ENABLE_TPL_EIGEN=OFF",
      "-DPRESSIO_ENABLE_TPL_PYBIND11=ON",
      "-DCMAKE_INSTALL_PREFIX={}".format(pressioPfxDir),
    ]
    if not os.path.exists(pressioBuildDir): os.makedirs(pressioBuildDir)
    subprocess.check_call(
      ["cmake", pressioSrcDir] + cmake_args, cwd=pressioBuildDir
    )
    subprocess.check_call(
      ["make", "install"], cwd=pressioBuildDir
    )
    pressioIncludeDir = self.build_temp + "/pressio/install/include"
    print("pressioIncludeDir = ", pressioIncludeDir)

    #------------------------
    # --- pressio4py ---
    #------------------------
    cfg = "Debug" if self.debug else "Release"

    # Set Python_EXECUTABLE instead if you use PYBIND11_FINDPYTHON
    # EXAMPLE_VERSION_INFO shows you how to pass a value into the C++ code
    # from Python.
    cmake_args = [
      "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={}".format(extdir),
      "-DPYTHON_EXECUTABLE={}".format(sys.executable),
      "-DCMAKE_BUILD_TYPE={}".format(cfg),
      "-DPRESSIO_INCLUDE_DIR={}".format(pressioIncludeDir),
      "-DVERSION_INFO={}".format(self.distribution.get_version()),
    ]
    build_args = []

    if "CXX" not in os.environ:
      msg = "CXX env var missing, needs to point to your target C++ compiler"
      raise RuntimeError(msg)

    # Set CMAKE_BUILD_PARALLEL_LEVEL to control the parallel build level
    # across all generators.
    if "CMAKE_BUILD_PARALLEL_LEVEL" not in os.environ:
      # for now, set parallel level to 4
      build_args += ["-j4"]
      # # self.parallel is a Python 3 only way to set parallel jobs by hand
      # # using -j in the build_ext call, not supported by pip or PyPA-build.
      # if hasattr(self, "parallel") and self.parallel:
      #   # CMake 3.12+ only.
      #   build_args += ["-j{}".format(self.parallel)]

    if not os.path.exists(self.build_temp):
      os.makedirs(self.build_temp)

    subprocess.check_call(
      ["cmake", ext.sourcedir] + cmake_args, cwd=self.build_temp
    )
    subprocess.check_call(
      ["cmake", "--build", "."] + build_args, cwd=self.build_temp
    )

setup(
  name="pressio4py",
  version="0.6.1rc3",
  author="Francesco Rizzi",
  author_email="fnrizzi@sandia.gov",
  description="pressio4py: projection-based model reduction for Python",
  #
  project_urls={
    'Documentation': 'https://pressio.github.io/pressio4py/html/index.html',
    'Source': 'https://github.com/Pressio/pressio4py'
  },
  #
  long_description=description(),
  ext_modules=[CMakeExtension("pressio4py")],
  cmdclass={"build_ext": CMakeBuild},
  install_requires=["numpy", "scipy", "matplotlib", "sklearn"],
  zip_safe=False,
  #
  python_requires='>=3',
  classifiers=[
    "License :: OSI Approved :: BSD License",
    "Operating System :: Unix",
    "Environment :: MacOS X",
    "Programming Language :: C++",
    "Programming Language :: Python :: 3 :: Only",
    "Topic :: Scientific/Engineering",
    "Topic :: Scientific/Engineering :: Mathematics",
    "Topic :: Scientific/Engineering :: Physics",
    "Topic :: Software Development :: Libraries",
    "Development Status :: 4 - Beta"
  ],
  keywords=["model reduction", "scientific computing", "dense linear algebra", "HPC"],
)




import numpy as np
import random

from pressio4py import rom as rom, logger

if __name__ == "__main__":
  '''
  A linear decoder (or mapping) represents

        y = phi * x

  where phi is the mapping's Jacobian (assumed constant for now).
  This is a typical approximation adopted in projection-based ROMs,
  where "x" is referred to as the generalized coordinates (or latent space)
  and "y" represents an approximation of the FOM state.

  This tutorial shows how to create a linear decoder/mapper in pressio4py.
  For examples using it, see subsequent tutorials.
  '''

  # create the matrix
  # attention: we declare phi to be column-major for these reasons:
  #
  # 1. pressio4py uses blas (wherever possible) to operate on numpy arrays,
  #    so a column-major layout implies seamless compatiblity with blas
  #
  # 2. when using column-major layout, pressio4py references the
  #    matrix phi without doing a deep copy, which saves memory
  #    since a single jacobian matrix is alive.
  #
  phi = np.ones((10,3), order='F')

  # to create the linear decoder, one can simply do
  linearDecoder = rom.Decoder(phi)

  # linearDecoder exposes a method to evaluate the mapping
  y,x = np.zeros(10), np.ones(3)
  linearDecoder.applyMapping(x, y)
  print(y)




import matplotlib.pyplot as plt

def edit_figure_for_web(ax, leg):
  mycolor = 'w'

  # makes all axes and text whie
  for l in ['bottom', 'left', 'right', 'top']:
    ax.spines[l].set_color(mycolor)

  ax.xaxis.label.set_color(mycolor);
  ax.tick_params(axis='x', colors=mycolor)
  ax.yaxis.label.set_color(mycolor);
  ax.tick_params(axis='y', colors=mycolor)

  for text in leg.get_texts(): text.set_color(mycolor)
